"""Notifications."""

from __future__ import annotations

from abilian.sbe.app import Application

# Constants
TOKEN_SERIALIZER_NAME = "unsubscribe_sbe"  # noqa: S105


def register_plugin(app: Application) -> None:
    cfg = app.config.setdefault("ABILIAN_SBE", {})
    cfg.setdefault("DAILY_SOCIAL_DIGEST_SUBJECT", "Des nouvelles de vos communautés")

    # TODO: Slightly confusing. Reorg?
    # from .tasks.social import DEFAULT_DIGEST_SCHEDULE, DIGEST_TASK_NAME
    from .views import notifications, social

    # CELERYBEAT_SCHEDULE = app.config.setdefault("CELERYBEAT_SCHEDULE", {})
    # if DIGEST_TASK_NAME not in CELERYBEAT_SCHEDULE:
    #     CELERYBEAT_SCHEDULE[DIGEST_TASK_NAME] = DEFAULT_DIGEST_SCHEDULE

    app.register_blueprint(notifications)



from __future__ import annotations

from typing import cast

from flask import g, request, url_for
from flask.blueprints import BlueprintSetupState
from flask_login import current_user

from abilian.i18n import _l
from abilian.services import get_service
from abilian.services.security import Admin, SecurityService
from abilian.web.action import Action, FAIcon, ModalActionMixin, actions


class ForumAction(Action):
    def is_current(self):
        return request.path == self.url()

    def is_filtered(self):
        filter_keys = ["today", "month", "year", "week"]
        current_url = request.path.split("/")
        filter = current_url[-1].strip()

        if filter not in filter_keys:
            return False

        return self.url() == "#filter"

    def url(self, context=None):
        if self._url or self.endpoint:
            return super().url(context=context)

        return url_for(f".{self.name}", community_id=g.community.slug)


class ThreadAction(ForumAction):
    def pre_condition(self, context):
        thread = actions.context.get("object")
        return bool(thread)


def is_admin(context):
    security = cast(SecurityService, get_service("security"))
    return security.has_role(current_user, Admin, object=context.get("object"))


def is_in_thread(context):
    thread = context.get("object")
    return not thread


def is_closed(context):
    thread = context.get("object")
    return thread.closed


def not_closed(context):
    return not is_closed(context)


class ForumModalAction(ModalActionMixin, ThreadAction):
    pass


_close_template_action = """
<form method="POST" action="{{ url }}" encoding="multipart/form-data">
  {{ csrf.field() }}
  <button type="submit" class="btn btn-link" name="action"
          value="{{ action.name}}">
    {%- if action.icon %}{{ action.icon }} {% endif %}
    {{ action.title }}
  </button>
</form>
"""

_actions = (
    ForumAction("forum:global", "index", _l("Recent conversations")),
    ForumAction("forum:global", "index/<string:filter>", _l("Top"), url="#filter"),
    ForumAction("forum:global", "archives", _l("Archives")),
    ForumAction(
        "forum:global", "attachments", _l("Attachments"), condition=is_in_thread
    ),
    ForumAction("forum:global", "new_thread", _l("New conversation"), icon="plus"),
    ForumModalAction(
        "forum:thread",
        "delete",
        _l("Delete"),
        condition=lambda ctx: is_admin(ctx) and not_closed(ctx),
        url="#modal-delete",
        icon="trash",
    ),
    #
    ThreadAction(
        "forum:thread",
        "close",
        _l("Close thread"),
        url="close",
        template_string=_close_template_action,
        condition=lambda ctx: is_admin(ctx) and not_closed(ctx),
        icon=FAIcon("lock"),
    ),
    ThreadAction(
        "forum:thread",
        "reopen",
        _l("Re-open thread"),
        url="close",
        template_string=_close_template_action,
        condition=lambda ctx: is_admin(ctx) and is_closed(ctx),
        icon=FAIcon("unlock"),
    ),
    ThreadAction(
        "forum:thread", "attachments", _l("Attachments"), url="attachments", icon="file"
    ),
)


def register_actions(state: BlueprintSetupState) -> None:
    if not actions.installed(state.app):
        return
    with state.app.app_context():
        actions.register(*_actions)



"""Extension for sending informations to upstream server."""

from __future__ import annotations

import typing
from typing import Any

from flask import Flask
from flask.globals import request_ctx
from flask.signals import request_finished, request_started
from flask.wrappers import Response

from abilian.core.signals import user_loaded

if typing.TYPE_CHECKING:
    from abilian.core.models.subjects import User


class UpstreamInfo:
    """Extension for sending informations to upstream server."""

    def __init__(self, app: Flask | None = None):
        if app is not None:
            self.init_app(app)

    def init_app(self, app: Flask):
        app.extensions["upstream"] = self
        request_started.connect(self.request_started, sender=app)
        request_finished.connect(self.request_finished, sender=app)
        user_loaded.connect(self.user_loaded, sender=app)

        config = app.config
        config.setdefault("ABILIAN_UPSTREAM_INFO_ENABLED", False)
        for key in ("ABILIAN_UPSTREAM_INFO_DISCARD", "ABILIAN_UPSTREAM_INFO_INCLUDE"):
            val = config.get(key, ())
            if val is not None:
                val = frozenset(val)
            config[key] = val

    def request_started(self, app: Flask):
        request_ctx.upstream_info = {"Username": "Anonymous"}

    def request_finished(self, app: Flask, response: Response):
        info = request_ctx.upstream_info
        config = app.config
        default_enabled = bool(config["ABILIAN_UPSTREAM_INFO_ENABLED"])
        disabled = config["ABILIAN_UPSTREAM_INFO_DISCARD"]
        enabled = config["ABILIAN_UPSTREAM_INFO_INCLUDE"]

        for key, val in info.items():
            if (default_enabled and key in disabled) or (
                not default_enabled and key not in enabled
            ):
                continue

            header = f"X-{key}"
            response.headers[header] = val

    def user_loaded(self, app: Flask, user: User, *args: Any, **kwargs: Any):
        request_ctx.upstream_info["Username"] = user.email


extension = UpstreamInfo()



from __future__ import annotations

from typing import TYPE_CHECKING

from flask_login import AnonymousUserMixin, LoginManager

if TYPE_CHECKING:
    from abilian.core.models.subjects import Group


class AnonymousUser(AnonymousUserMixin):
    def has_role(self, role):
        from abilian.services import get_security_service

        security = get_security_service()
        return security.has_role(self, role)

    @property
    def groups(self) -> set[Group]:
        return set()


login_manager = LoginManager()
login_manager.anonymous_user = AnonymousUser



""""""

# from __future__ import annotations

# import contextlib

# import sqlalchemy as sa

# from abilian.core.models.tag import Tag, entity_tag_tbl
# from abilian.i18n import _

# # from ..search.criterion import BaseCriterion
# from .extension import ENTITY_DEFAULT_NS_ATTR


# class TagCriterion(BaseCriterion):
#     """Filter entities with selected tag(s)."""

#     form_default_value = ""

#     def __init__(self, *args, **kwargs):
#         if len(args) == 0:
#             kwargs.setdefault("name", "tags")

#         if len(args) < 2:
#             kwargs.setdefault("label", _("Tags"))

#         super().__init__(*args, **kwargs)

#     @BaseCriterion.model.setter
#     def model(self, model):
#         BaseCriterion.model.fset(self, model)
#         self.ns = getattr(model, ENTITY_DEFAULT_NS_ATTR)

#     @property
#     def valid_tags(self):
#         join_clause = entity_tag_tbl.join(
#             self.model, self.model.id == entity_tag_tbl.c.entity_id
#         )
#         model_tags = sa.sql.select([entity_tag_tbl.c.tag_id], from_obj=join_clause)
#         return Tag.query.filter(Tag.ns == self.ns, Tag.id.in_(model_tags)).all()

#     def get_request_values(self, request):
#         tag_ids = []

#         for val in request.values.getlist(self.name):
#             with contextlib.suppress(ValueError):
#                 tag_ids.append(int(val))

#         valid_tags = set(self.valid_tags)
#         tags = []
#         for tid in tag_ids:
#             t = Tag.query.get(tid)
#             if t in valid_tags:
#                 tags.append(t)

#         return tags

#     def filter(self, query, module, request, searched_text, *args, **kwargs):
#         tags = self.get_request_values(request)
#         if not tags:
#             return query

#         cond = sa.sql.exists(
#             sa.sql.select(
#                 [1],
#                 sa.sql.and_(
#                     entity_tag_tbl.c.entity_id == self.model.id,
#                     entity_tag_tbl.c.tag_id.in_(t.id for t in tags),
#                 ),
#             )
#         )
#         return query.filter(cond)

#     @property
#     def form_filter_type(self):
#         return "select"

#     @property
#     def form_unset_value(self):
#         return []

#     @property
#     def form_filter_args(self):
#         # expected value: [list of selectable items, is multiple?]
#         return [[(str(t.id), t.label) for t in self.valid_tags], True]



"""Blob.

References to files stored in a on-disk repository
"""

from __future__ import annotations

import hashlib
import uuid
from pathlib import Path
from typing import IO

import sqlalchemy as sa
from sqlalchemy.event import listens_for
from sqlalchemy.orm.session import Session
from sqlalchemy.orm.unitofwork import UOWTransaction
from sqlalchemy.schema import Column
from sqlalchemy.types import Integer

from abilian.core.models.base import Model
from abilian.core.sqlalchemy import UUID, JSONDict


class Blob(Model):
    """Model for storing large file content.

    Files are stored on-disk, named after their uuid. Repository is
    located in instance folder/data/files.
    """

    __tablename__ = "blob"

    id = Column(Integer(), primary_key=True, autoincrement=True)
    uuid = Column(UUID(), unique=True, nullable=False, default=uuid.uuid4)
    meta = Column(JSONDict(), nullable=False, default=dict)

    def __init__(self, value=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.uuid is None:
            self.uuid = uuid.uuid4()

        if self.meta is None:
            self.meta = {}

        if value is not None:
            self.value = value

    def _md5_hexdigest(self) -> str:
        """Return md5.hexdigest() of self.value."""
        if isinstance(self.value, str):
            content = self.value.encode("utf-8")
        else:
            content = self.value
        return hashlib.md5(content).hexdigest()  # noqa: S324

    @property
    def file(self) -> Path | None:
        """Return :class:`pathlib.Path` object used for storing value."""
        from abilian.services.blob_store import session_blob_store

        return session_blob_store.get(self, self.uuid)

    @property
    def size(self) -> int:
        """Return size in bytes of value."""
        file = self.file
        if file is None:
            return 0
        return file.stat().st_size

    @property
    def value(self) -> bytes | None:
        """Binary value content."""
        file = self.file
        if file is None:
            return None
        return file.read_bytes()

    @value.setter
    def value(self, value: bytes | str | IO):
        """Store binary content to applications's repository and update
        `self.meta['md5']`.

        :param:content: bytes, or any object with a `read()` method
        :param:encoding: encoding to use when content is Unicode
        """
        from abilian.services.blob_store import session_blob_store

        session_blob_store.set(self, self.uuid, value)

        if self.value:
            self.meta["md5"] = self._md5_hexdigest()

        filename = getattr(value, "filename", None)
        if filename:
            if isinstance(filename, bytes):
                filename = filename.decode("utf-8")
            self.meta["filename"] = filename

        content_type = getattr(value, "content_type", None)
        if content_type:
            self.meta["mimetype"] = content_type

    @value.deleter
    def value(self):
        """Remove value from repository."""
        from abilian.services.blob_store import session_blob_store

        session_blob_store.delete(self, self.uuid)

    @property
    def md5(self) -> str | None:
        """Return md5 from meta, or compute it if absent."""
        md5 = self.meta.get("md5")
        if md5 is None and self.value:
            md5 = self._md5_hexdigest()

        return md5

    def __bool__(self) -> bool:
        """A blob is considered falsy if it has no file."""
        return self.file is not None and self.file.exists()

    # Py3k compat
    __nonzero__ = __bool__


@listens_for(sa.orm.Session, "after_flush")
def _blob_propagate_delete_content(session: Session, flush_context: UOWTransaction):
    deleted = (obj for obj in session.deleted if isinstance(obj, Blob))
    for blob in deleted:
        del blob.value
